/* 
!! Purpose:<br>
!!-------------------------------------------------------------------------<br>
!!`Aov` is a tool for period analysis of irregularly sampled time series, 
!!such as stellar and planetary transit light curves. It relies on state-of-art
!!statistically poweful, robust and fast algorithms implemented in
!!my `AovPer` package, partly based on my original research and
!!publications. All routines come in two flavours, with/without weights.<br>
!!References:<br>
!!--------------------------------------------------------------------------<br>
!!The methods used are based on my 3 papers, listed in the
!!corresponding routine headers:<br>
!!aov-phase binning aov (Schwarzenberg-Czerny, A., 1989, MNRAS 241, 153);<br>
!!aovmh-multi-harmonic aov (Schwarzenberg-Czerny, A., 1996, ApJ(Lett) 460L, 107);<br>
!!aovtr-transit/eclipse aov (Schwarzenberg-Czerny, A. & Beaulieu, J.-Ph., 
!!2006, MNRAS 365, 165);<br>
!!These papers and my software are also available from my web page
!!`http://www.camk.edu.pl/~alex/`<br>
!!Please cite this papers accordingly in publications using the current package.<br>
!!Input:<br>
!!--------------------------------------------------------------------------<br>
!!Observations - one ASCII file per star, 2 or 3 columns in order: time,value,error<br> 
!!`aov.par` - file containing default values of program parameters<br>
!!The file `aov.par` must exist, at least with dummy content `&par /`. 
!!Aov command line: may reset parameters from aov.par<br>
!!Star name is either in STAR parameter or input from file named in STRLST.
!!File of star observations is named DIR+STAR+EXT, where + marks concatenation.<br>
!!Running `./aov STAR=SIMUL` generates sample input (see Usage).
!!Compilation:<br>
!!---------------------------------------------------------------------------<br>
!! AOV is compiled simply by:
!!
!! gcc -o aov AovSub.c AovObs.c AovSpec.c AovPer.c Aov.c -lm
!! or
!! gcc -ggdb -o aov AovSub.c AovParms.c AovObs.c AovSpec.c AovPer.c Aov.c -lm -fsanitize=address -fno-omit-frame-pointer
!!Usage:<br>
!!-------------------------------------------------------------------------<br>
!!Running `./aov` generates sample input parameter file `aov.par` and
!!data file `SIMUL1.dat`. Then command `./aov STAR=SIMUL1` executes them in the 
!!ordinary way as if user star name was SIMUL1.<br>
!!Program Parameters:
!!---------------------------------------------------------------------<br>
!!Obligatory Parameters (listed either in aov.par or command line):
!!`  character*(20)    :: STAR=""!  star name or ...   `<br>
!!`  character*(20)    :: STRLST=""! name of file of star names  `<br>
!!<br>
!!Optional command parameters of main interest and their default values`<br>
!!`  real(TIME)        :: FR0=0. !  lower frequency`<br>
!!`  real(TIME)        :: FRU=4. !  upper frequency`<br>
!!`  real(TIME)        :: FRS=0. !  frequency step (0.-optimum sampling)`<br>
!!`                       !  warning: undersampling risk for FRS>0. `<br>
!!`  integer           :: NPAR=5  ! number of model parameters`<br>
!!`  integer           :: METHOD=AOVMH ! select from`<br>
!!`                          ! AOV,AOVTR,AOVMH,AOVW,AOVTRW,AOVMHW`<br>
!!`                          ! Those ending 'W' process weights.`<br>
!!`  integer           :: NCOV=2 ! number of bin coverages`<br>
!!`  character*(20)    :: DIR=''     ! Light curve file name prefix`<br>
!!`  character*(20)    :: EXT='.dat' ! Light curve file name extension`<br>
!!Full list is in header of `mprms` module.<br>

!!Input parameter NPAR sets number of phase bins NPAR phase bins or 
!!NPAR/2 Fourier harmonics.
!!I was using 3<=NPAR<=9 and I do not recommend exceeding these limits
!!in general period surveys. For transient search use much higher values 
!!of NPAR, so that transit width ~1/NPAR. In theory, the CPU time
!!is proportional to ~NPAR, but as NPAR increases the spectral
!!resolution also increases, so it is advisable to decrease
!!the frequency step by 1/NPAR (done automatically by the code),
!!which causes CPU time scaling like ~NPAR^2. There
!!are possible additional CPU savings, but these are not
!!worth worrying about at this stage. NPAR=5 worked well
!!for near sinusoidal oscillations.<br>
!!FSAMPL needs to be set =>3. The FSAMPL=4 choice means 4 frequencies covering 
!!a spectral line. This is in my opinion close to optimal as the code fits 
!!line profile to find peak frequency. If speed is no problem one can use e.g. FSAMPL=7.<br>
!!The fastest are `AOV` or `AOVW`, depending on `ERRIN`, yet they
!!suffer from an artificial noise/residuals due to sharp steps between phase bins.
!!`AMH` and `AMHW` employ a smooth Fourier model hence are statistically better, yet
!!about 10 times slower. All four algorithms complexity scales proportional
!!to number of observations, frequencies and model parameters.<br>

!!Methods:<br>
!!------------------------------------------------------------------<br>
!!The algorithms employ either bins and sub-bins or projection onto
!!ortogonal polynomials yielding as fast implementation as possible when
!!irregular sampling with large gaps make FFT impractical.
!!Use of orthogonal functions, either phase binning or 
!!trigonometric polynomials results in statistically near optimal
!!behaviour in that cross-talk/correlation between model elements is minimal.
!!Because of that the algorithms remain fairly robust against outliers. 
!!More importantly use of orthogonal functions permits 
!!application of the classical Fisher AnoVa statistics to
!!evaluate frequency spectrum and reliability of any detected periods. 
!!Depending on `SPTYPE` the spectrum obeys either
!!Fisher-Snedecor F or beta probability distributions.<br>

!!Experience:<br>
!!-----------------------------------------------------------------<br>
!!The original periodogram/frequency spectrum routines were used on
!!over 10^5 stars, as part of f95 TATRY package in OGLE and LCVSS
!!mass photometry surveys and in EROS c package 
!!All Star Searcher, among others used to analyse 318 light curves
!!of suspected variable stars in the field M31C. For the Cepheids
!!and the eclipsing binaries. Tests against periods set by manual 
!!inspection of periodograms and light curves were performed 
!!on 25 000 light curves from ASAS South. The comparision involved the most
!!significant periods only. The periods always 
!!agreed or were 2om or 3om or om/2 sub-/harmonics of the published frequency om.
!!For Cepheids NPAR=5 worked well. For eclipsing binaries with proximity 
!!effects (e.g. W UMA) or for sinusoidal variables NPAR=3 was better,
!!however 2om ambiguities occured.<br>
!!Trials demonstrated that the aov detection criterion
!!used in the present code works better than the Laefler-Kinman,
!!PDM and Stetson criteria for eclipsing or pulsating variable stars.
!!However, picking right NPAR for your problem requires some
!!trial runs before decision on the final NPAR. For some applications,
!!such as transit search it sis recommended to perform several runs,
!!e.g. for NPAR=15, 30 and 60, to scan a range of transit widths.<br>

!! History:<br>
!!-------------------------------------------------------------------<br>
!! 1989 - development of AOV phase binning routines;<br>
!! 1996 - AOVMH Fourier series routines added;<br> 
!! 1999-2002 - all AOV routines expanded for use as module in TATRY code;<br>
!! 2003 - porting of AOV routines into C for use in EROS and DIRECT projects.
!! This resulted in major upgrade/factor several sppeding of the algorithm AOV by 
!! introduction of subbins;<br>
!! 2004 - new transient search algorithm fitting top-hat/box function;<br>
!! 2018 - subsstantial revision: main changes were in array declarations and 
!!normalization of observations. All fixed lengh arrays are replaced with the
!!dynamic ones. In periodogram subroutines observations are now copied into local
!!arrays when needed. Input observations are left intact. Small inconsistencies 
!!were removed and AOV and AOVTR routines were combined into one AOV routine
!!performing both ordinary phase folding and binning or transit search. 
!!F95 and C packages became compatible to a high degree. Auto documentation using `ford`.
!!Phases are now calculated in a uniform way using:<br>
!!`dph=fr*t[i]`<br>
!!`dph-=floor(dph)`<br>
!!Tests on AMD PC and gcc demonstrated that `-=floor` is faster than `fmod`.<br>
!! 2020 - improved handling of command line string parameters, introduction of
!! optional list of star names.<br> 
!!
!!Copyrights:<br>
!!The whole package is subject to copyrights by its author, 
!!(C) A. Schwarzenberg-Czerny 1089-2020,  alex@camk.edu.pl 
!!It is free to use and copy, yet distribution of modiffications or fragments 
!!is forbidden. It is unsuitable for application carrying risk of damages.
!!To acknowledge its use, pleas cite papers listed in the References section.<br> !  */

#include "Aov.h"

INT main(INT argc, char *argv[]) {

/* Variables ---------------------------------------------------*/ 
    SPEC_T spec;
    OBS_T obs;                               
    FILE *strres, *strls, *parm;
    INT simul, list;
    LONG ifr;
    INT metcnt, io, ib, i;
    TIME fb[NBEST], tsav;
    FLOAT *f=NULL;
    FLOAT thm, dfm, qual;
    FLOAT sav;

    char fname[STRING], *s;

/* Executable body ---------------------------------------------*/
    printf("\n   AOV Periodogram Routines\n    Ver. %s\n",VERSION);
    printf("   (C) Alex Schwarzenberg-Czerny alex@camk.edu.pl\n\n");

/* Set default string parameters */
    ini_parm();

/* Get parameters from the parameter file */
    io=1;
    i=0;
    while ((s=get_parm("aov.par"))!=NULL) {
      io = io && (set_parm(s) != 0);
      i++;
    }
    io=(i==0) || io;
/* Get parameters from the command line */
    while(--argc) io = io && (set_parm(*(++argv)) != 0);

    if(io==0) { /* wrong aov.par or command syntax*/
      printf("Usage:\n"
         "./aov to generate sample .par and .dat files and next\n"
         "./aov STAR=SIMUL1 to run those .par and .dat files\n"
         "./aov  [[option=value]...] for general use, where\n"
         "option=FR0|FRU|FRS|NH2|METHOD|SPTYPE|NCOV|MFRQ|DIR|EXT|STAR|STRLST\n"
         "METHOD=aov|atr|amh|psp|aovw|atrw|amhw|pspw\n"
         "SPTYPE=RAW|AOV|MODEL|RESID\n"
         "EPOCH=MIN|AVE|MED|number\n"
         "Option names are not case sensitive\n"
         "Values must not contain ''='' sign\n"
         "String values are case sensitive and bracket "" is needed\n"
         "when they contain SPACE or special characters\n"
         "File aov.par has f90 namelist format\n"
         "Command line parameter values have precedence over aov.par\n"
         "If STRLST=fname, star names are input from file fname\n"
         "aov.par syntax is similar, except commas as separators and\n" 
         "&par start and / end keys.\n");
      exit(EXIT_FAILURE); 
    }

    NCOV=MAX(1,NCOV); /* Print parameters */
    FRU=MAX(FR0,FRU);
    NPAR=MAX(2,NPAR);
    printf("Parameters:\n");
    prt_parms(stdout);

    /* Check STRLST file */
    strls = fopen (STRLST, "r");
    list = (strls != NULL);
    if (!list && strlen(STRLST)>0) {
      printf (" Failed to open STRLST=%s\n", STRLST);
      printf (" Instead Tries STAR=%s\n", STAR);
    };

/*--------------------------------------------------------------------*/
    do { /* STAR loop */
      if (list) { /* Get star names from STRLST file, if it exists */
        do {
          io=fscanf(strls, "%40s", STAR);
        } while (io>0);
        if (io==EOF) {
          fclose(strls);
          break;
        };
      };

    /* get data ...*/
    simul = strcmp (strup(msg,STAR), "SIMUL")==0;
    if (simul) {
        /* ...from simulation or ... */
        obs = test_obs (100);
        strcpy(STAR,"SIMUL1");
        parm=fopen("aov.par","w");
        prt_parms(parm);
        fclose(parm);       
    } else {
        /* ... from a file. NOTE: appends file prefix/extension, if any */
        strcpy (fname, DIR);
        strcat (fname, STAR);
        strcat (fname, EXT);
        obs=get_obs(fname);
        if (obs.no<=0) {
          printf("***No valid data for STAR=%s\n",STAR);
          if (list) continue;
          break;
        }
    }
    /* process observations */
    if (obs.no <= NPAR)
        printf ("Too few observations, abandoning %s\n", STAR);
    else {
/* Set up frequency grid and get periodogram */
        tsav = obs.t[1]-obs.t[0];
        for (i=1;i<(INT)(sizeof(obs.t)/sizeof(TIME));i++) 
          tsav=MAX(sav,obs.t[i]-obs.t[i-1]);
        if ( tsav < 0. )  
          printf("***Warning: time not sorted in %s\n",STAR);
        printf("Are errors given? %d\n",obs.weights);
        metcnt=0;
        spec=fgrid (obs.no, obs.t, NPAR/2, FR0, FRU, FRS);
        if ( (f = malloc (spec.nfr*sizeof (FLOAT) ) ) == NULL) {
            printf ("Failed memory allocation: f");
            continue;
        };
        for (ifr = 0; ifr < spec.nfr; ifr++)
            f[ifr] = (FLOAT)( spec.frs * ifr + spec.fr0);
        if ( (obs.tmax - obs.tmin) * spec.frs > 0.3 && (VERBOSE) )
            printf (" AOV: warning: undersampling in frequency\n");
//        for (method=0;method<8;method++) {
            if (!metcnt) printf("grid: %15lg %15lg %ld\n",
                spec.fr0,spec.frs,spec.nfr);
/* Calculate periodogramme */
            io = AovDrv (METHOD, SPTYPE, obs.no, obs.t, obs.v, obs.w,
                    NPAR, NCOV, spec);

/* Write result periodogram into file */
            strcpy (fname, STAR);
            strcat (fname, ".res");
            strres = fopen (fname, "w");
            if (strres == NULL) {
                printf (" Failed to open %s\n", fname);
                continue;
            };
            for (ifr=0; ifr<spec.nfr; ifr++) 
                fprintf (strres,"%15g %10.5f\n",f[ifr],log10(spec.th[ifr]));
            fclose (strres);

/* Find & remove peaks */
            for (ib = 0; ib < NBEST; ib++) {
                peakrm(obs.tmax - obs.tmin, spec, &fb[ib], &thm, &dfm);
                if (ib == 0) qual = (FLOAT) log10 (thm);
            };
/* Print summary */
            if(!metcnt) printf ("meth star  no_obs quality best_frequencies\n");
            printf ("%4.4s %6s%5i%8.3f", met_code(METHOD,msg), 
                STAR, obs.no, qual);
            for (ib = 0; ib < NBEST; ib++)
                printf ("%10.6f", fb[ib]);
            printf ("\n");
            metcnt++;
//        };
        free (f); clr_spec(spec);
    };
    clr_obs(obs);
    } while (list); /* end STAR loop */

    printf ("AOV finished\n\n");
    return (0);
};





