/*
!   Routines for pre-processing of observations.
!   This is part of the AOV package of routines for computation and 
!   evaluation of AOV periodograms.
!   This package is subject to copyrights by its author, 
!   (C) A. Schwarzenberg-Czerny 1998-2018,  alex@camk.edu.pl 
!   Its distribution is free, except that distribution of modifications is
!   prohibited.
 */

#include "Aov.h"

/*-----------------------------------------------------------*/

OBS_T cre_obs(INT no, INT weights){
// ! create structure for observations
  OBS_T obs = {no, 1, 0., 0., 0., NULL, NULL, NULL};
   obs.t = malloc(no*sizeof(TIME));
   obs.v = malloc(no*sizeof(FLOAT));
   if (weights) obs.w = malloc(no*sizeof(FLOAT));
   return obs;
}

void clr_obs(OBS_T obs) {
// ! clear structure for observations
   if (obs.w != NULL) free(obs.w);
   free(obs.v); free(obs.t);
}

void norm_times (char epstr[], OBS_T *obs) {
/*
!! Purpose: <br>
!! Normalize times prior period analysis. Depending on EPSTR and 
!! presence of weights calculates EPOCH and subtracts it from 
!! all times. This should be done for numerical reasons so that 
!! times span zero. Since user may wish a fixed epoch, e.g. to
!! use consistent phase bins, this is not performed inside 
!! periodogram routines.

! EPSTR: selects epoch: 'MIN','AVE','MED' or real number string
*/

    char *ch, eptab[]={"min ave med"}; // or number
    INT epmod, epsize=strlen(eptab)/4;
    INT j, *ind;
    FLOAT sav, *tfl;
    TIME epval;

    ind=malloc((*obs).no*sizeof(INT));
    tfl=malloc((*obs).no*sizeof(FLOAT));
    for (j=0; j<(*obs).no; j++) tfl[j]=(FLOAT)(*obs).t[j];
    sortx_m ((*obs).no, tfl, ind);
    free(tfl); 
    (*obs).tmin=(*obs).t[ind[0]];
    (*obs).tmax=(*obs).t[ind[(*obs).no-1]];

    epmod=epsize+1;     // check epstr in eptab
    ch=strstr(eptab,epstr);
    if(ch!=NULL) epmod=(ch-eptab)/4;
    epval=(TIME)0.;
    if (epmod>epsize) { // read epoch value from epstr
       j=sscanf(epstr,"%lf",&epval);
       if (j!=1) {
          if (VERBOSE) printf("norm_obs: wrong EPOCH string\n");
          exit(8);
       } else if((epval-(*obs).tmin)*(epval-(*obs).tmax)>0. && VERBOSE) 
          printf("norm_times: warning: EPOCH out of times range");
    } else {   
       switch(epmod) {  // normalize time
          case 0: 
             epval=(*obs).tmin;
             break;
          case 1:
             epval=0.;
             sav=0.;
             for (j=0; j<(*obs).no; j++) if ((*obs).w[j]>0.) {
                epval+=(*obs).t[j]*(*obs).w[j];
                sav+=(*obs).w[j];
             };
             epval/=sav;
             break;
          case 2: 
             epval=(*obs).t[ind[(*obs).no/2]];
             break;
          default:
             if (VERBOSE) printf("norm_obs: wrong epmod\n");
             exit(7);
             break;
          }
     }
     if (VERBOSE) printf("EPOCH= %f\n",epval);
     (*obs).epoch=epval;
     for (j = 0; j < (*obs).no; j++) 
        (*obs).t[j]-=epval;

     for (j = 1; j < (*obs).no; j++) 
        if((*obs).t[j]-(*obs).t[j-1]<(TIME)0.) break;
     if (j<((*obs).no) && (VERBOSE) ) printf("norm_obs: warning: times not sorted\n");
     free(ind);
}

OBS_T get_obs(char *fname) {
/*
Reads 2 or 3 ascii columns of observations from file fname.
Returns structure obs storing them. Number of columns is set
basing on first line containing 2 or 3 float numbers. Lines 
inconsistent with the first one are skipped.
*/
   OBS_T obs;
   typedef struct node {
      TIME t;
      FLOAT v, w;
      struct node *next;
   } NODE_T;
   NODE_T *cur, *sav, *first;
   FILE *in;
   INT i, num, cnt=0, cols=0, skip=0;
   char *buffer;
   size_t bufsize = 80;

   obs.no = 0;

   buffer = (char *)malloc(bufsize * sizeof(char));   
   in=fopen(fname,"r"); /* get file */
   if (in == NULL) {
      printf ("get_obs: Failed to open %s\n", fname);
      return obs;
   };
   cur=malloc(sizeof(NODE_T));
   first=cur;

   while ((num=getline(&buffer,&bufsize,in))>0) { /* read one line */
      num=sscanf(buffer,"%lg %g %g", &((*cur).t),&((*cur).v),&((*cur).w));
      if(num>0) {
         if (cnt==0 && num>=2 && num<=3) cols=num;
         if (num==cols) {
            cnt++;
            sav=malloc(sizeof(NODE_T));
            if (cols==2) (*cur).w=1.;
            (*cur).next=sav;
            (*sav).next=NULL;
            cur=sav;
         } else skip++;
      } else skip++;
   }
   fclose(in);
   free(buffer);
   if (VERBOSE) printf("Accepted %d, Skipped %d\n",cnt,skip);

/* copy observations into arrays */
   obs = cre_obs(cnt, 1);
   if (cnt<=0) return obs;
   obs.weights=(cols==3);
   cur=first; 
   cnt=0;
   while ((*cur).next!=NULL) {
      obs.t[cnt]=(*cur).t;
      obs.v[cnt]=(*cur).v;
      obs.w[cnt++]=(*cur).w;
      sav=cur;
      cur=(*cur).next; 
      free(sav);
   }
   free(cur);
// normalize times and weights
// For numerical reasons subtract epoch from times 
   norm_times(EPOCH,&obs);
   for (i=0;i<cnt;i++) 
      if(obs.w[i]!=0.) obs.w[i]=1./(fabs(obs.w[i])*obs.w[i]);
   return obs;
}


OBS_T test_obs (INT no) {
    INT k, j, *ind;
    FLOAT s2n, span, omt, *tfl;
    TIME fr, dph;
    OBS_T obs;
    FILE *strlc;
    /* simulate data */
    obs=cre_obs(no,1);
    s2n = (FLOAT) 100.;
    span = (FLOAT) 40.;
    fr = 1. / (TIME) 0.3617869;
    printf (" Simulated data, FR0=%15.7f\n", fr);
    for (k = 0; k < no; k++) {
        obs.w[k] = (FLOAT) 0.5;
        obs.t[k] = sin (1. * k);
        obs.t[k] = span * obs.t[k] * obs.t[k];
        obs.t[0] = 0.;
        obs.t[obs.no - 1] = span;
        dph = fr * obs.t[k];
        omt = PI2 * (FLOAT) (dph - floor (dph) );
        obs.v[k] = (FLOAT) (11.90 + 0.434 * cos (omt + 4.72) +
                        0.237 * cos (2. * omt + 0.741) +
                        sin (1.7 * k) / s2n);
    };
    obs.weights=1;
    strlc = fopen ("SIMUL1.dat", "w");
    if (strlc == NULL) {
        printf (" Failed to open %s\n", "SIMUL1.dat");
        exit (2);
    };

    ind=malloc(obs.no*sizeof(INT));
    tfl=malloc(obs.no*sizeof(FLOAT));
    for (k=0; k<obs.no; k++) tfl[k]=(FLOAT)obs.t[k];
    sortx_m (obs.no, tfl, ind);
    for (k=0; k<obs.no; k++) {
        j=ind[k];
        fprintf (strlc, "%f %f %f\n", obs.t[j],obs.v[j],1./sqrt(obs.w[j]));
    }
    fclose (strlc); 
    free(tfl); free(ind);
    norm_times(EPOCH,&obs);
    return obs;
};


