/*
 AovParms.c
Handling parameter values
 */
#include "Aov.h"

/* parameters and their default values */
    char msg[STRING]; /* temporary string space, not for saving */
/* Default parameter values */
    TIME FR0=(TIME)0.; /*  lower frequency */
    TIME FRU=(TIME)4.; /*  upper frequency */
    TIME FRS=(TIME)0.; /*  upper frequency */
    INT NPAR=5;         /* number of model parameters */
    INT NCOV=2;        /* number of bin coverages */
    INT MXFRQ=1000000;        /* max number of frequencies */
    enum METHOD_T METHOD=AMH; /* ID of periodogram method */
    enum SPTYPE_T SPTYPE=SP_AOV; /* ID of periodogram form */
    char STAR[STRING]; /*star name*/ 
    char STRLST[STRING]; /*star list file*/ 
    char DIR[STRING];     /* light curve file name prefix */
    char EXT[STRING];   /* light curve file name extension */

/* keywords dictionaries searched by dict_val. dict_key functions */
static const char *opt_dic="fr0 fru frs npar method sptype ncov mxfrq "
  "star strlst ext dir";
static const char *met_dic="aov atr amh psp aovw atrw amhw pspw"; 
static const char *spt_dic="raw aov mod res";
/*-----------------------------------------------------------*/
void ini_parm() {
/* initialize defaults of string values */
    strcpy(STAR,"SIMUL"); /* special value to start simulation run */
    strcpy(STRLST,""); 
    strcpy(DIR,"");     
    strcpy(EXT,".dat"); 
}
/*-----------------------------------------------------------*/
static INT dict_val(const char *dict, char *key) {
/* Returns key position in dictionary, 0 if not found */
   char *t; 
   INT i=0,j;
   t=strstr(dict,key);
   if (t!=NULL) {
     for (j=(INT)(t-dict);j>=0;j--) if(dict[j]==' ') i++;
     return ++i;
   } else return 0;
}
/*-----------------------------------------------------------*/
char *dict_key(const char *dict, INT no, char*key) {
/* returns pointer to dictionary keyword number no (starting from 1) or
  NULL pointer for error input */
   INT i=0,j,k;
   no--;
   k=0;
   for (j=0;j<(INT)strlen(dict);j++) {
     if(dict[j]==' ') { /* find start and end of keyword */
       i++;
       if(no==i) k=j+1;
       if(no<i) break;
     }
   }
   if (i<no) return NULL;
   for (i=j-k;--i>=0;key[i]=dict[k+i]); /* copy keyword */
   key[j-k]='\0';
   return key;
}
char *met_code(INT met, char*msg) {
/* a wrapper so that met_dic is not accessed from outside of this file */
  return dict_key(met_dic,met,msg);
}
/*-----------------------------------------------------------*/
void prt_parms(FILE *str) {
/* writes all parameters in F90 NML file format, for input in future runs */
  char spt[6],met[6];
  dict_key(met_dic,METHOD,met); trim(met);
  dict_key(spt_dic,SPTYPE,spt); trim(spt);
  fprintf(str,"&par\nFR0=%g,\nFRU=%g,\nFRS=%g,\nNPAR=%d,\n"
     "METHOD=\"%s\",\nSPTYPE=\"%s\",\nNCOV=%d,\nMXFRQ=%d,\n"
     "STAR=\"%s\",\nSTRLST=\"%s\",\nEXT=\"%s\",\nDIR=\"%s\"\n/\n",
     FR0,FRU,FRS,NPAR,met,spt,NCOV,MXFRQ,STAR,STRLST,EXT,DIR);
}
/*-----------------------------------------------------------*/
char *get_parm(char *fname) {
/* Routine returns consecutive tokens from F90 NML format parameter file.
   At first call it reads the file 'fname' and makes table of pointers to 
   its tokens. No array keywords are permitted and tokens are separated with 
   commas.
   Note: between calls the vital info is saved in static variables. */

  #define LFILE 2000 /* max length of parameter file, bytes */
  #define MXPAR 20   /* max number of different tokens */
  static char *str, *parm[MXPAR];
  static FILE *st=NULL;
  static INT no,l;
  char *t, *s;
  INT i,j;

  if (st==NULL) { /* executes at first call */
    st=fopen(fname,"r");
    if (st==NULL) {
      printf("%s%s\n",fname," parameter file missing");
      return NULL;
    } else { /* Allocate memory and read the whole par file */
      str = malloc(LFILE);
      no = fread(str,1,LFILE,st);
      fclose(st); 
      str[no]='\0';
    /* Check F90 NML file format of str, optional for other applictions ...*/
      j=0;l=0;
      for (i=0;i<no;i++) {
        if(str[i]=='=') j++;
        if(str[i]==',') l++;
      }
      trim(str);
      t=strrchr(str,'/');
      s=strtok(str," \t\n"); /* NOTE: no changes of str are permitted between
                                calls of strtok, otherwise it fails. */
    /* ...up to here check F90 NML*/
      if (j != l+1 || strcmp(s,"&par") !=0 || t==NULL) { /*test NML syntax */
        printf("%s%s%s\n","Wrong syntax of ",fname,", ignored");
        return NULL;
      } else { /* Read and trim from blanks keywords separated by comma, 
                  no NML suntax tested here. */
        strcpy(t,"");
        l=0;
        do {
          s=strtok(NULL,",");
          if (s!=NULL) {
            i=strspn(s," \n\t");
            if (i<(INT)strlen(s)) {
              parm[l]=s+i;
              l++;
            }
          } 
        } while (s!=NULL);
        trim(parm[l-1]);
        for (i=l;i<MXPAR;i++) parm[i]=NULL;
      }
      no=0;
    }
  } /* end of first call initialization */
  if(no==l) free(str); /* just return next pointer */
  return parm[no++];
}
/*-----------------------------------------------------------*/
INT set_parm(char *arg) {
/* Routine takes one string keyword and sets value of the corresponding 
   parameter. No SPACE nor COMMA discriminators are permitted in it.
   No array parameters are permitted. The valid syntax is name=value.
   If value is a string, it may be surrounded with quotation marks "..."*/
  INT opt_no, io;
  char *opt_nam, *opt_val;
      io=1;
      /* get parameter name, its code number and string containing its value */
         opt_nam=strtok(arg,"="); /* no changes in arg permitted between strtok
                                     calls */
         opt_val=strtok(NULL,",");
         strlow(opt_nam,opt_nam);
         opt_no=dict_val(opt_dic, opt_nam);
//printf("OPT= %s %d %d %s\n",opt_nam,opt_no,(INT)strlen(opt_val),opt_val);
         switch(opt_no) { /* implement option */
            case 1: /* numerical parameters */
              io=sscanf(opt_val,"%lf",&FR0);
              break;
            case 2:
              io=sscanf(opt_val,"%lf",&FRU);
              break;
            case 3:
              io=sscanf(opt_val,"%lf",&FRS);
              break;
            case 4:
              io=sscanf(opt_val,"%d",&NPAR);
              break;
            case 5: /* get METHOD code*/
              strip(opt_val);
              strlow(opt_val,opt_val);
              METHOD=dict_val(met_dic, opt_val);
              if(METHOD==ERRM) io=0;
              break;
            case 6: /* get SPTYPE code*/
              strip(opt_val);
              strlow(opt_val,opt_val);
              SPTYPE=dict_val(spt_dic, opt_val);
              if(SPTYPE==ERRS) io=0;
              break;
            case 7: /* numerical parameters */
              io=sscanf(opt_val,"%d",&NCOV);
              break;
            case 8:
              io=sscanf(opt_val,"%d",&MXFRQ);
              break;
            case 9: /* general string parameters from here */
              io=0;
              if(strip(opt_val)==NULL) break;              
              strncpy(STAR,opt_val,STRING);
              io=1;
              break;
            case 10:
              strip(opt_val);              
              strncpy(STRLST,opt_val,STRING);
              break;
            case 11:
              strip(opt_val);              
              strncpy(EXT,opt_val,STRING);
              break;
            case 12:
              strip(opt_val);              
              strncpy(DIR,opt_val,STRING);
              break;
            default: /* Set general error flag, see METHOD and SPTYPE too.*/ 
              io=0;
              break;
          }
  if (!io) printf("wrong option/value: %s %s\n",strup(msg,opt_nam),opt_val);
  return io; /* logical error flag */
}

