/* AovPer.c

  Purpose: 
  State-of-art routines for period search in unevenly sampled time series.
  Package contains routines for phase folding and binning, Fourier series fit
  power spectrum and search of planetary transits, for weighted and unweighted
  observations.

  Calling:
  Demo calling example is contained in AOVDRV routine in this file.
  Prior to analysis all observations need to be normalized, e.g. by call
  of NORM_OBS routine. Normalization involves time, value and weights pre-
  processing. Values and weight arrays (obs.v and obs.w) may become destroyed 
  after spectrum routine call. All routines share common calling sequence, yet
  depending on routine some parameters are ignored.
  INT sptype - determines type of frequency spectrum returned from routine;
           for SP_RAW,SP_AOV,SP_MODEL,SP_RESID the spectrum obeys: 
           none, F, beta and another beta probability distributions
  OBS_T obs - input structure containing arrays of normalized observations 
          (time, value and weight)
  FLOAT sw, FLOAT vf - input sum of weights and weighted sum of squares
          of normalized values. Ones are used for no explicit weights.
  INT npar - input number of parameters of mathematical model in use;
  INT ncov - input number of different phase origins employed to smooth
          effects of time discretization
  SPEC_T spec - on input contains parameters of spectrum frequency grid
          (nfr, fr0, frs where freq=ifr*frs+fr0, ifr=0,...,nfr-1);
          on output spec.th contains resultant frequency spectrum.

  Methods:
  The Aov package contains state-of-art routines for period search and
  planetary transit search by phase folding and binning of light curves 
  and/or by multi-harmonic Fourier series fit. For a particular set up
  these routines reproduce popular box-least squares and generalised Lomb 
  Scargle algorithm, ableit in more efficient implementations involving sub bins
  or orthogonal projections. All routines come in two flavours: for
  weighted and unweighted observations. Their calling sequences are uniform,
  though depending on routine some parameters are ignored.

  The period routines split into two groups: 
  (i) AOV/TR & AOVMH calculate quadratic norm of a fitted model. 
  As such their results obey Parceval theorem. Depending on SPTYPE_T parameter 
  the result spectrum may be converted to classical statistics: either aov 
  (Fisher F) spectrum or the corresponding beta statistics spectra of normalized
  'model power' or 'residual chi2' flavour.
  (i.A) AOV/TR: Depending on sign of npar: for npar>0 aov/tr routine fits 
  folded light curve with a histogram of npar bins, for npar<0 aov/tr routine 
  fits light curve with just two phase bins: in- and out-of suspect transit. 
  The transit width is fixed by npar while its position is fine-tuned 
  depending on ncov. (Reference: A.Schwarzenberg-Czerny, 1989, MNRAS 241, 153;
  A.Schwarzenberg-Czerny, & J.-Ph. Beaulieu,  2006, MNRAS 365, 165.
  (i.B) AOVMH fits Fourier series of npar terms by orthogonal projection on 
  Szego polynomials. Reference: A.Schwarzenberg-Czerny, 1996, ApJ 460L, 107.
  In particular for npar=3 AOVMH returns Generalised 
  Lomb-Scargle spectrum of Ferraz-Mello, 1981 and Zechmeister & Kurster, 2009.
 
  (ii) The other two routines, PWSP & PWSPW, return squared amplitude of a fitted 
  sinusoid. The underlying algorithms differ from each other even for unary
  weights. For uneven sampling the result power spectra obey no Parceval theorem
  hence power spectra obey no classical probability distribution. Provided all
  input values are 1 PWSP returns window function (PWSPW does not).
 
  Several general utility routines are provided: 
  NORM_OBS to normalize observations, GRID, PEAK and PEAKRM to set up frequency
  spectrum and analyse result peaks.
  AOVDRV to call spectrum routines and SPECTRUM to convert results into
  frequency spectra. Note that arrays obs.v and obs.w (values and weights)
  may be destroyed during computations.

  This package is subject to copyrights by its author,
  (C) A. Schwarzenberg-Czerny 1998-2018, alex@camk.edu.pl
  Its distribution is free, except that distribution of modifications is
  prohibited. Please acknowlege use by citation of papers listed in the
  corresponding routines.
 */

#include "Aov.h"

/*-----------------------------------------------------------*/

static void spectrum(enum SPTYPE_T sptype, INT no, INT npar, FLOAT vf, 
     SPEC_T spec) {
/* Purpose: Converts quadratic norm of fitted model into
   one of classic frequency spectrum statistics
*/
   LONG l; 
   FLOAT fac;
   switch (sptype) { //SP_RAW,SP_AOV,SP_MODEL,SP_RESID
   case SP_RAW: // do nothing
       break;
   case SP_AOV: // convert to F(npar,no-npar;th)
       {fac=(no-npar)/(npar-1.);
        for (l=0;l<spec.nfr;l++) 
          spec.th[l]=spec.th[l]/MAX(vf-spec.th[l],TINY)*fac;
       }
       break;
   case SP_MODEL: // convert to ibeta(npar/2,(no-npar)/2;th)
       {for (l=0;l<spec.nfr;l++) 
          spec.th[l]=spec.th[l]/vf;
       }
       break;
   case SP_RESID: // convert to ibeta((no-npar)/2,npar/2;th)
       {for (l=0;l<spec.nfr;l++) 
          spec.th[l]=1.-spec.th[l]/vf;
       }
       break;
   default:
       {if (VERBOSE) printf("Spectrum: wrong SPTYPE_T\n");}
       break;
   }

};

/* ==========================================================================
! AOVMH & AOVMHW routines for multiharmonic Fourier spectrum

!   Purpose: 
!   Respectively for no weights and weights present AOVMH & AOVMHW  fit
!   unevenlysampled time series with a multiharmonic Fourier series and  
!   return spectrum for a range of Fourier frequencies. On input NPAR 
!   sets number of Fourier terrms (NPAR/2 harmonics). For NPAR=3 and
!   SPTYPE_T=SP_MODEL, AOVMH & AOVMHW results are identical to 
!   Ferraz-Mello(1981) and Zechmeister & Kurster (2009) generalized 
!   Lomb-Scargle (GLS) spectrum. NCOV is ignored.
!
!   Method: 
!   By Szego recurrence AOVMH & AOVMHW generate orthogonal trigonometric 
!   polynomials and next by orthogonal projection expand uneven sampled time 
!   series in terms of these polynomials. Such an algorithm constitutes the
!   fastest implementation of Fourier series least squares fit. For NPAR>>3
!   these algorithms are particularly suitable to detect smooth non-sinusoidal
!   oscillations involving sharp features (e.g. eclipses and narrow pulses).
!   For detection of sine signals in statistical sense (sensitivity) AOVMH &
!   AOVMHW never perform worse than either Lomb-Scargle or Power Spectrum. 
!   For non-sinusoidal signals AOVMH & AOVMHW advantage becomes significant. 
!   For identical NPAR, AOVMH & AOVMHW are about 10 times slower than more crude
!   AOV/AOVW routines. 
 
!   Copyrights and Distribution:
!   This routines are subject to copyrights by its author, 
!   (C) A. Schwarzenberg-Czerny 1998-2005,  alex@camk.edu.pl 
!   Its distribution is free, except that distribution of modifications 
!   is prohibited.
!   Please quote A.Schwarzenberg-Czerny, 1996, Astrophys. J.,460, L107
!   while using AOVMH & AOVMHW. 
*/

static INT aovmh (enum SPTYPE_T sptype,
       INT no, TIME *t, FLOAT *vin, FLOAT *win,
       INT npar, INT ncov, SPEC_T spec) {
//! Routine for no weights
     
    INT l, n, k, nn2, nh;
    FLOAT sn, s, sav, v2;
    FLOAT *v, *cfr, *cfi, *pr, *pi, *zr, *zi, *znr, *zni;
    FLOAT sr, si, alr, ali, scr, sci;
    TIME fr, dph;

    l = (win==NULL)+ncov; /* dummy use */
    nh=MAX(npar/2,1);
    nn2 = nh + nh;


// normalize observations
    v=malloc(no*sizeof(FLOAT));
    sav=0.;
    for (k = 0; k < no; k++) sav+=vin[k];
    sav/=(FLOAT)no;
    v2=0.;
    for (k = 0; k < no; k++) {
        v[k]=vin[k]-sav;
        v2+=v[k]*v[k];
    }
    if ( (no < nn2 + 2) ) {
        if (VERBOSE) fprintf (stderr, " AOVMH: error: wrong size of arrays\n");
        return (1);
    };
    cfr=malloc(no*sizeof(FLOAT));
    cfi=malloc(no*sizeof(FLOAT));
    pr= malloc(no*sizeof(FLOAT));
    pi= malloc(no*sizeof(FLOAT));
    zr= malloc(no*sizeof(FLOAT));
    zi= malloc(no*sizeof(FLOAT));
    znr=malloc(no*sizeof(FLOAT));
    zni=malloc(no*sizeof(FLOAT));

    for (l = 0; l < spec.nfr; l++) {
        spec.th[l] = (FLOAT) 0.;
        fr = spec.frs * l + spec.fr0;
        for (k = 0; k < no; k++) {
            dph = fr * t[k];
            dph-= floor(dph);
            sav = PI2 * (FLOAT) dph;
            zr[k] = (FLOAT) cos (sav);
            zi[k] = (FLOAT) sin (sav);
            pr[k] = znr[k] = (FLOAT) 1.;
            pi[k] = zni[k] = (FLOAT) 0.;
            sav = sav * nh;
            cfr[k] = v[k] * (FLOAT) cos (sav);
            cfi[k] = v[k] * (FLOAT) sin (sav);
        }
        for (n = 0; n <= nn2; n++) {
            sn = alr = ali = scr = sci = (FLOAT) 0.;
            for (k = 0; k < no; k++) {
                sn += (pr[k] * pr[k] + pi[k] * pi[k]);
                /* new formula for alpha */
                alr += zr[k] * pr[k] - zi[k] * pi[k];
                ali += zr[k] * pi[k] + zi[k] * pr[k];
                scr += (pr[k] * cfr[k] + pi[k] * cfi[k]);
                sci += (pr[k] * cfi[k] - pi[k] * cfr[k]);
            }
            sn = MAX (sn, EPSILON);
            alr /= sn;
            ali /= sn;
            spec.th[l] += (scr * scr + sci * sci) / sn;
            for (k = 0; k < no; k++) {
                sr = alr * znr[k] - ali * zni[k];
                si = alr * zni[k] + ali * znr[k];
                s = pr[k] * zr[k] - pi[k] * zi[k] 
                        - sr * pr[k] - si * pi[k];
                pi[k] = pr[k] * zi[k] + pi[k] * zr[k] 
                        + sr * pi[k] - si * pr[k];
                pr[k] = s;
                s = znr[k] * zr[k] - zni[k] * zi[k];
                zni[k] = zni[k] * zr[k] + znr[k] * zi[k];
                znr[k] = s;
            }
        }
    };
    spectrum(sptype, no, nn2+1, v2, spec);

    free(zni); free(znr); free(zi);  free(zr); free(pi);  free(pr); 
    free(cfi); free(cfr);free(v);
    return (0);
}

static INT aovmhw (enum SPTYPE_T sptype,
       INT no, TIME *t, FLOAT *vin, FLOAT *win,
       INT npar, INT ncov, SPEC_T spec) {
//! Routine for weighted observations
     
    INT n, nn2, nh, noeff;
    INT l, k;
    FLOAT sn, s, sav, v2;
    FLOAT *rw, *vr, *cfr, *cfi, *pr, *pi, *zr, *zi, *znr, *zni;
    FLOAT sr, si, alr, ali, scr, sci;
    TIME fr, dph;

    l = ncov; /* dummy use */
    nh=MAX(npar/2,1);
    nn2 = nh + nh;
// normalize observations
    vr=malloc(no*sizeof(FLOAT));
    rw=malloc(no*sizeof(FLOAT));
    sav=0.;
    v2=0;
    noeff=0;
    for (k = 0; k < no; k++) 
       if (win[k]>0.) { 
          rw[k]=sqrt(win[k]/4.);
          sav+=vin[k]*win[k]/4.;
          v2+=win[k]/4.;
          noeff+=1;
       } else rw[k]=0.;
    sav/=v2;
    v2=0.;
    for (k = 0; k < no; k++) {
       vr[k]=(vin[k]-sav)*rw[k];
       v2+=vr[k]*vr[k];
    };
    if ( (noeff < nn2 + 2) ) {
        if (VERBOSE) fprintf (stderr, " AOVMH: error: wrong size of arrays\n");
        return (1);
    };
    cfr=malloc(no*sizeof(FLOAT));
    cfi=malloc(no*sizeof(FLOAT));
    pr= malloc(no*sizeof(FLOAT));
    pi= malloc(no*sizeof(FLOAT));
    zr= malloc(no*sizeof(FLOAT));
    zi= malloc(no*sizeof(FLOAT));
    znr=malloc(no*sizeof(FLOAT));
    zni=malloc(no*sizeof(FLOAT));

    for (l = 0; l < spec.nfr; l++) {
        spec.th[l] = (FLOAT) 0.;
        fr = spec.frs * l + spec.fr0;
        for (k = 0; k < no; k++) {
            dph = fr * t[k];
            dph-= floor (dph);
            sav = PI2 * (FLOAT) dph;
            zr[k] = (FLOAT) cos (sav);
            zi[k] = (FLOAT) sin (sav);
            znr[k] = (FLOAT) 1.;
            pr[k] = rw[k];
            pi[k] = zni[k] = (FLOAT) 0.;
            sav = sav * nh;
            cfr[k] = vr[k] * (FLOAT) cos (sav);
            cfi[k] = vr[k] * (FLOAT) sin (sav);
        }
        for (n = 0; n <= nn2; n++) {
            sn = alr = ali = scr = sci = (FLOAT) 0.;
            for (k = 0; k < no; k++) {
                sn += (pr[k] * pr[k] + pi[k] * pi[k]);
                /* new formula for alpha */
                alr += (zr[k] * pr[k] - zi[k] * pi[k]) * rw[k];
                ali += (zr[k] * pi[k] + zi[k] * pr[k]) * rw[k];
                scr += (pr[k] * cfr[k] + pi[k] * cfi[k]);
                sci += (pr[k] * cfi[k] - pi[k] * cfr[k]);
            }
            sn = MAX (sn, EPSILON);
            alr /= sn;
            ali /= sn;
            spec.th[l] += (scr * scr + sci * sci) / sn;
            for (k = 0; k < no; k++) {
                sr = alr * znr[k] - ali * zni[k];
                si = alr * zni[k] + ali * znr[k];
                s = pr[k] * zr[k] - pi[k] * zi[k] 
                      - sr * pr[k] - si * pi[k];
                pi[k] = pr[k] * zi[k] + pi[k] * zr[k] 
                      + sr * pi[k] - si * pr[k];
                pr[k] = s;
                s = znr[k] * zr[k] - zni[k] * zi[k];
                zni[k] = zni[k] * zr[k] + znr[k] * zi[k];
                znr[k] = s;
            }
        }
    };
    spectrum(sptype, noeff, nn2+1, v2, spec);
    free(zni); free(znr); free(zi); free(zr); free(pi); free(pr); 
    free(cfi); free(cfr); free(rw); free(vr);  
    return (0);
}
//  End of AOVMH & AOVMHW routines for multiharmonic Fourier spectrum
// ==========================================================================
/* AOV & AOVW routines for phase folding and binning and/or transit period search

! Purpose:
! The state-of-art routines for either period search by phase folding and 
! binning of observations (for NP>0) or for search for periodic transits/pulses 
! of short duty cycle, with constat signal elsewhere (for NP<0). In the latter 
! case data are fitted with just two unequal phase bins: in- and out-of transit.

! Method:
! In any case observations are phase folded and binned into NP equal phase
! bins starting at NCOV sub-bin phases. Subsequently the algorithm 
! either sums squares of bin averages (for NP>0). For transit search
! (NP<0) algorithm searches for the maximum value phase bin (considered 
! in-transit for magnitude units). Its squared value and identities stemming 
! from Parceval theorem are employed to evaluate sum of squares for two
! phase-bins in- and out-of transit. Note: here transits are defined as MAXIMA
! (i.e. in magnitude units). For flux units (minima) reverse sign of observations.
! If algorithm detects poor phase coverage (bins with <CTMIN observations)
! it rebins observations into bins of near equal occupancy, by sorting
! observations in phase. This is a rare situation hence amounts to no
! significant slowing down. Thanks to this feature, for null frequency
! these routines return useful variability indicator as long as the sort 
! routine is stable (e.g. merge sort).
!
! Advantage:
! By checking and repairing phase coverage these routines perform robust and
! near-optimal period search, returnind spectrum of known classical 
! probability distribution, equally well for small and large data. Slight
! statistical inefficiency stems from step-like function yielding slightly
! worse fit that equal size Fourier fit. This may be remedied by calling
! AOVMH/AOVMHV instead, at the speed cost. In terms of speed the present 
! AOV/AOVW implementation is state-of-art as it employs pre-binning into 
! NP*NCOV sub-bins and only afterwords combines subbins into suitably 
! phased bins. See B. Enoch et al., 2012, A&A 548, A48 
! for independent evaluation of the present sub-bin algorithm.
   
! Copyrights and Distribution:
! This package is subject to copyrights by its author, 
! (C) A. Schwarzenberg-Czerny 2003-2018,  alex@camk.edu.pl 
! Its distribution is free, except that distribution of modifications
! is prohibited.
! Please quote A. Schwarzenberg-Czerny & J.-Ph. Beaulieu, 2005, MNRAS 365, 165,
! while using AOV or AOVW.
*/
static INT aov (enum SPTYPE_T sptype,
      INT no, TIME *t, FLOAT *vin, FLOAT *w,
      INT npar, INT ncov, SPEC_T spec) {
//! Routine for no weights
    INT ibin, ip, nbc, transit;
    INT *ncnt, *ind, i;
    LONG ifr, iflex;
    FLOAT *ph, *v, *ave, sav, v2;
    TIME fr, dph;
   
    i=(NULL==w); /* dummy use*/

    /* Set variables (incl. mean and variance) */
    transit = npar < 0;
    if (transit) npar = -npar;
// normalize observations
    v=malloc(no*sizeof(FLOAT));
    sav=0.;
    for (i = 0; i < no; i++) sav+=vin[i]; 
    sav/=(FLOAT)no;
    v2=0.;
    for (i = 0; i < no; i++) { 
        v[i]=vin[i]-sav;
        v2+=v[i]*v[i];
    };
    if ( (no <= npar+npar) ) {      
        if (VERBOSE)  fprintf(stderr,"AOV: error: wrong size of arrays\n");
        return (-1);
    };
    nbc = npar * ncov;
    ncnt=malloc((nbc+ncov)*sizeof(INT));
    ave=malloc((nbc+ncov)*sizeof(FLOAT));
    ind=malloc(no*sizeof(INT));
    ph= malloc(no*sizeof(FLOAT));
    iflex = 0;
    /*
     * Loop over all frequencies
     */
    for (ifr = 0; ifr < spec.nfr; ifr++) {
        fr = ( ( (TIME) ifr) * spec.frs + spec.fr0);
        /*
         * Up to two passes over all frequencies, depending on phase
         * coverage
         */
        for (ip = 0; ip < 2; ip++) {
            for (ibin = 0; ibin < nbc+ncov; ibin++) {
                ave[ibin] = (FLOAT) 0.;
                ncnt[ibin] = 0;
            };
            if (ip == 0)       /* Default fixed size phase bins */
                for (i = 0; i < no; i++) {
                    dph = t[i] * fr;
                    dph-= floor (dph);
                    ph[i] = (FLOAT) dph;
                    ibin = (INT) (dph * nbc);
                    ave[ibin] += v[i];
                    ++ncnt[ibin];
                }
            else                  /* For poor phase coverage optional
                                 * flexible bins */
            {
                /*
                 * sort index ind using key ph0, any index sort
                 * routine would do
                 */
                ++iflex;
                sortx_m (no, ph, ind);
                for (i = 0; i < no; i++) {
                    ibin = i * nbc / no;
                    ave[ibin] += v[ind[i]];
                    ++ncnt[ibin];
                };
            };
            /*
             * counts: sub-bins=>bins
             */
            for (ibin = 0; ibin < ncov; ibin++)
                ncnt[ibin + nbc] = ncnt[ibin];
            i = 0;
            for (ibin = ncov + nbc - 1; ibin >= 0; ibin--)
                ncnt[ibin] = (i += ncnt[ibin]);
            for (ibin = 0; ibin < nbc; ibin++)
                ncnt[ibin] -= ncnt[ibin + ncov];
            for (ibin = 0; ibin < nbc; ibin++)
                if (ncnt[ibin] < CTMIN)   /* check bin occupation */
                    break;
            if (ibin >= nbc)
                break;
        };
        /*
         * data: sub-bins=>bins
         */
        for (ibin = 0; ibin < ncov; ibin++)
            ave[ibin + nbc] = ave[ibin];
        sav = (FLOAT) 0.;
        for (ibin = ncov + nbc - 1; ibin >= 0; ibin--)
            ave[ibin] = (sav += ave[ibin]);
        for (ibin = 0; ibin < nbc; ibin++)
            ave[ibin] -= ave[ibin + ncov];

        if (transit) {
        /*
         * AoV statistics for transits
         */
    	    sav = ave[0] / ncnt[0];
    	    for (ibin = 0; ibin < nbc; ibin++)
    	        if ( (ave[ibin] /= ncnt[ibin]) >= sav) {
    	            sav = ave[ibin];
    	            i = ibin;
    	        };
    	    sav *= sav * ncnt[i] * no / (FLOAT) (no - ncnt[i]);
        } else {                         /* where 'vf' keeps the total sum of
                                 * squares of f[] */
        /*
         * Calculate A.O.V. statistics for a given frequency
         */
    	    sav = (FLOAT) 0.;
     	    for (ibin = 0; ibin < nbc; ibin++)
     	        sav += (ave[ibin] * ave[ibin] / ncnt[ibin]);
     	    sav /= (FLOAT) ncov;
        }
        spec.th[ifr]=sav;
    };
    if (VERBOSE && iflex > 0) fprintf(stderr,"AOV:warning: poor phase"
       " coverage at %ld frequencies\n",iflex);
    spectrum(sptype, no, transit ? 2 : npar, v2, spec);
    free(ph); free(ind); free(ave); free(ncnt); free(v);
    return (0);
};

static INT aovw (enum SPTYPE_T sptype,
       INT no, TIME *t, FLOAT *vin, FLOAT *win,
       INT npar, INT ncov, SPEC_T spec) {
//! Routine for weighted observations
    INT ibin, ip, nbc, transit; 
    INT *ind, i, noeff;
    LONG ifr, iflex;
    FLOAT *ph, *v, *w, *ncnt, *ave, sav, ctw, sw, v2;
    TIME fr, dph;
    /*
     * Set variables (incl. mean and variance)
     */
    transit = npar < 0;
    if (transit) npar = -npar;

// normalize observations
    v= malloc(no*sizeof(FLOAT));  
    w= malloc(no*sizeof(FLOAT));
    noeff=0;
    sw=0.;
    ctw=0.;  
    for (i = 0; i < no; i++) 
       if (win[i]>0.) { 
          w[i]=win[i];
          sw+=w[i];
          ctw+=vin[i]*w[i];
          noeff+=1;
       } else w[i]=0.;
    ctw/=sw;
    v2=0.;
    for (i = 0; i < no; i++) {
       v[i]=vin[i]-ctw;
       v2+=w[i]*v[i]*v[i];
       v[i]*=w[i];
    };
    if ( (noeff <= npar + npar) || (noeff < CTMIN * npar) ) {
        if (VERBOSE) fprintf(stderr,"AOVw: error: wrong size of arrays/n");
        return (-1);
    };
    nbc = npar * ncov;
    ncnt=malloc((nbc+ncov)*sizeof(FLOAT));
    ave=malloc((nbc+ncov)*sizeof(FLOAT));
    ind=malloc(no*sizeof(INT));
    ph= malloc(no*sizeof(FLOAT));
    /*
     * calculate totals and normalize variables
     */
    iflex = 0;
    ctw = ( (FLOAT) CTMIN) * sw / ( (FLOAT) noeff);
    /*
     * Loop over all frequencies
     */
    for (ifr = 0; ifr < spec.nfr; ifr++) {
        fr = ( ( (TIME) ifr) * spec.frs + spec.fr0);
        /*
         * Up to two passes over all frequencies, depending on phase
         * coverage
         */
        for (ip = 0; ip < 2; ip++) {
            for (ibin = 0; ibin < nbc; ibin++) {
                ave[ibin] = ncnt[ibin] = (FLOAT) 0.;
            };
            if (ip == 0)       /* Default fixed size phase bins */
                for (i = 0; i < no; i++) {
                    dph = t[i] * fr; 
                    dph-= floor (dph);
                    ph[i] = (FLOAT) dph;
                    ibin = (INT) (dph * nbc);
                    ave[ibin] += v[i];
                    ncnt[ibin] += w[i];
                }
            else                  /* For poor phase coverage optional
                                 * flexible bins */
            {
                /*
                 * sort index ind using key ph0, any index sort
                 * routine would do
                 */
                ++iflex;
                sortx_m (no, ph, ind);
                for (i = 0; i < no; i++) {
                    ibin = i * nbc / no;
                    ave[ibin] += v[ind[i]];
                    ncnt[ibin] += w[ind[i]];
                };
            };
            /*
             * counts: sub-bins=>bins
             */
            for (ibin = 0; ibin < ncov; ibin++)
                ncnt[ibin + nbc] = ncnt[ibin];
            sav = (FLOAT) 0.;
            for (ibin = ncov + nbc - 1; ibin >= 0; ibin--)
                ncnt[ibin] = (sav += ncnt[ibin]);
            for (ibin = 0; ibin < nbc; ibin++)
                ncnt[ibin] -= ncnt[ibin + ncov];
            for (ibin = 0; ibin < nbc; ibin++)
                if (ncnt[ibin] < ctw)
                    break;
            if (ibin >= nbc)
                break;
        };
        /*
         * data: sub-bins=>bins
         */
        for (ibin = 0; ibin < ncov; ibin++)
            ave[ibin + nbc] = ave[ibin];
        sav = (FLOAT) 0.;
        for (ibin = ncov + nbc - 1; ibin >= 0; ibin--)
            ave[ibin] = (sav += ave[ibin]);
        for (ibin = 0; ibin < nbc; ibin++)
            ave[ibin] -= ave[ibin + ncov];

        if (transit) {
        /*
         * AoV statistics for transits
         */
 	    sav = ave[0] / ncnt[0];
            for (ibin = 0; ibin < nbc; ibin++)
                if ( (ave[ibin] /= ncnt[ibin]) >= sav) {
                    sav = ave[ibin];
                    i = ibin;
                };
            sav *= sav * ncnt[i] * sw / (FLOAT) (sw - ncnt[i]);
        } else {
        /*
         * Calculate A.O.V. statistics for a given frequency
         */
            sav = (FLOAT) 0.;
            for (ibin = 0; ibin < nbc; ibin++)
                sav += (ave[ibin] * ave[ibin] / ncnt[ibin]);
            sav /= (FLOAT) ncov;
        }
        spec.th[ifr]=sav;
    };
    if (VERBOSE && iflex > 0) fprintf(stderr,"AOVw:warning: poor phase"
       " coverage at %ld frequencies\n",iflex);
    spectrum(sptype, noeff, transit ? 2 : npar, v2, spec);
    free(ph);free(ind);free(ave);free(ncnt);free(w);free(v);
    return (0);
};

//! End of AOV & AOVW routines
//!==========================================================================
//! POWSP & POWSPW routines returning two different versions of power spectrum

static INT powsp (enum SPTYPE_T sptype,
        INT no, TIME *t, FLOAT *vin, FLOAT *w,
        INT npar, INT ncov, SPEC_T spec) {
    /*
!   Purpose:
!   POWSP - returns discrete power spectrum (DPS) for unevenly sampled
!   observations. For vin==1 calculates window function (WF). 
!   SPTYPE_T, NPAR and NCOV are ignored. The normalization factor is 1/NOBS**2,
!   suitable for WF. Power is defined in a usual way as sum of squares of
!   projections onto sine and cosine functions. This concept does not extend
!   on weighted observations, hence POWSPW employs different algorithm.
!   Reference: T.Deeming, 1975, Ap&SS 36, 137.
!
!  Copyrights and Distribution:
!  This routine is subject to copyrights by its author, 
!  (C) A. Schwarzenberg-Czerny 1998-2018,  alex@camk.edu.pl 
!  Its distribution is free, except that distribution of modifications is
!  prohibited.
     */
    LONG l;
    INT k;
    FLOAT *v, cfr, cfi, fac, sav;
    TIME fr, dph;

    k=(INT)sptype+(NULL==w)+npar+ncov; /* dummy use */

// normalize observations
    v=malloc(no*sizeof(FLOAT));
    sav=0.;
    for (k = 0; k < no; k++) sav+=vin[k];
    sav/=no;
    for (k = 0; k < no; k++) v[k]=vin[k]-sav;
    if ( (no < 4) ) {
        if (VERBOSE) fprintf (stderr, " POWSP: error: wrong size of arrays\n");
        return (1);
    };

    fac = (FLOAT) no;
    fac = 1./(fac*fac);
    for (l = 0; l < spec.nfr; l++) {
        cfr = cfi = (FLOAT) 0.;
        fr = spec.frs * l + spec.fr0;
        for (k = 0; k < no; k++) {
            dph = fr * t[k];
            dph-= floor (dph);
            sav = PI2 * (FLOAT) dph;
            cfr += v[k] * (FLOAT) cos (sav);
            cfi += v[k] * (FLOAT) sin (sav);
        }
        spec.th[l] = (cfr*cfr+cfi*cfi)*fac;
    }
    free(v);
    return (0);
}

static INT powspw (enum SPTYPE_T sptype,  
        INT no, TIME *t, FLOAT *vin, FLOAT *win,
        INT npar, INT ncov, SPEC_T spec) {
    /*
!   Purpose:
!   POWSPW - returns discrete power spectrum (DPS) for weighted unevenly sampled
!   observations defined differently than in POWSP. No window function (WF) is
!   available and NPAR and NCOV are ignored. The algorithm by weighted least
!   squares fits data with Asin(om*t)+Bcos(om*t) and for each frequency returns
!   A**2+B**2. The normalization is peculiar to this implementation and depends
!   on time sampling.
!   Reference: S. T. Fletcher et al., 2011, MNRAS 415, 1310 (see Appendix).
!
!  Copyrights and Distribution:
!  This routine is subject to copyrights by its author, 
!  (C) A. Schwarzenberg-Czerny 2018,  alex@camk.edu.pl 
!  Its distribution is free, except that distribution of modifications is
!  prohibited.
     */
    INT k, noeff;
    LONG l;
    FLOAT *v, *w, s, c, sx, cx, ss, cc, sc, sav;
    TIME fr, dph;

    k=(INT)sptype+npar+ncov; /* dummy use */

// normalize observations
    v=malloc(no*sizeof(FLOAT));
    w=malloc(no*sizeof(FLOAT));
    noeff=0;
    c=0.;
    s=0.; 
    noeff=0; 
    for (k = 0; k < no; k++) 
       if (win[k]>0.) { 
          w[k]=win[k];
          s+=w[k];
          c+=vin[k]*w[k];
          noeff+=1;
       } else w[k]=0.;
    c/=s;
    for (k = 0; k < no; k++) v[k]=(vin[k]-c)*w[k];
    if ( (no < 4) ) {
        if (VERBOSE) fprintf (stderr, " POWSPW: error: wrong size of arrays\n");
        return (1);
    };

    for (l = 0; l < spec.nfr; l++) {
        sx=cx=ss=cc=sc = (FLOAT) 0.;
        fr = spec.frs * l + spec.fr0;
        for (k = 0; k < no; k++) {
            dph = fr * t[k];
            dph-= floor (dph);
            sav = PI2 * (FLOAT) dph;
            c=(FLOAT) cos (sav);
            s=(FLOAT) sin (sav);
            sx+=s*v[k];
            cx+=c*v[k];
            ss+=w[k]*s*s;
            cc+=w[k]*c*c;
            sc+=w[k]*s*c;
        }
        spec.th[l] = 0.25*(pow(sx*cc-cx*sc,2.)+pow(cx*ss-sx*sc,2.))/
            MAX(pow(ss*cc-sc*sc,2.),TINY);
    }
    free(w); free(v);
    return (0);
}
// End of POWSP & POWSPW routines 
//========================================================
// AOVDRV driver routine for the frequency spectrum routines

INT AovDrv (enum METHOD_T method, enum SPTYPE_T sptype, 
        INT nin, TIME tin[], FLOAT vin[], FLOAT win[],
        INT npar, INT ncov, SPEC_T spec) {
    /*Purpose: example calling of spectrum routines.

    method,sptype - type of method and result spectrum
    t(:) - times
    v(:) - values
    w(:) - weights or errors, depending on ERRIN
    npar,ncov - number of parameters and phase coverages
    spec - input: frequency spectrum grid.
           output: spec.th spectrum
    */
    INT i;

    /* get spectrum/periodogram */
    switch (method) {// AOV,ATR,AMH,PWS,AOVW,ATRW,AMHW,PWSW
    case AOV:
        i = aov (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    case ATR:
        i = aov (sptype, nin, tin,vin,win,-MAX (2, npar), ncov, spec);
        break;
    case AMH:
        i = aovmh (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    case PWS:
        i = powsp (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    case AOVW:
        i = aovw (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    case ATRW:
        i = aovw (sptype, nin, tin,vin,win,-MAX (2, npar), ncov, spec);
        break;
    case AMHW:
         i = aovmhw (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    case PWSW:
        i = powspw (sptype, nin, tin,vin,win, MAX (2, npar), ncov, spec);
        break;
    default:
        if (VERBOSE) fprintf (stderr, "Aov: error: Wrong method");
        return (1);
    };
    return (i);
};


