/*
! Routines for handling frequency spectra.
! This file is part of AOV package of routines for computation and evaluation 
! of periodograms.
! (C) A. Schwarzenberg-Czerny 1998-2018, alex@camk.edu.pl
! Its distribution is free, except that distribution of modifications is
! prohibited.
 */

#include "Aov.h"

/*-----------------------------------------------------------*/

SPEC_T cre_spec(LONG nfr, TIME fr0, TIME frs) {
   SPEC_T spec = {nfr, fr0, frs, NULL};
   spec.th = malloc(nfr*sizeof(FLOAT));
   if (spec.th == NULL) {
      printf ("Failed memory allocation: th");
      exit (1);
   };
   return spec;
}

void clr_spec(SPEC_T spec) {
   free(spec.th);
}

SPEC_T fgrid (INT n, TIME t[], INT npar, TIME fr0, TIME fru, TIME frs) {
/*
! Purpose:
! Robust routine for evaluation of the time sampling pattern
! and guessing a suitable frequency grid. The spec structure containing 
! grid parameters is returned as function value.

! Input:
! N,T[N] number and values of times
! NPAR - number of model parameters (to refine frequency grid)
! FR0 - start frequency
! FRS,FRU - if positive set grid values, though fru>=frs is forced;
!           otherwise they are calculated within routine.
! Output: SPEC structure:
! spec.NFR, spec.FR0 and spec.FRS
! where NFR is calculated as (FRU-FR0)/FRS+1

! (C) A. Schwarzenberg-Czerny 1998-2018, alex@camk.edu.pl
! Distribution of FGRID is free, except that distribution of modifications is
! prohibited.
 */
    INT i, *ind;
    LONG nf;
    TIME t0;
    FLOAT *d;

    ind=malloc(n*sizeof(INT));
    d=malloc(n*sizeof(FLOAT));

    fru=MAX(fru,frs);
    if(frs<=(TIME)0.) {
      t0=(t[n-1]+t[0])*0.5;
      for (i=0;i<n;i++) d[i]=(FLOAT)(t[i]-t0);
      sortx_m(n,d,ind);
      frs=t[ind[nint(0.93*(n-1))]]-t[ind[nint(0.07*(n-1))]];
      frs = 1. / (frs * npar * FSAMPL);
      if(fru<=(TIME)0.) {
        for (i=1;i<n;i++) d[i-1]=(FLOAT)(t[ind[i]]-t[ind[i-1]]);
        sortx_m(n-1,d,ind);
        fru=(TIME)1./d[ind[nint(0.33*(n-2))]];
      }
    }
    fru=MAX(fru,fr0);
    nf = nint( (fru - fr0) / (frs) ) + 1;
    if (nf > MXFRQ) {
      nf = MXFRQ;
      frs = (fru-fr0)/(nf-1);
      printf(
"FGRID: *** DANGER *** Data spacing too fine or span too long for good sampling\n" 
"of periodogramme. Brute solution to increase MXFRQ costs extra execution time.\n"
"You may choose to preset shorter or coarser frequency grid by first loading   \n"
"shorter span data to choose a narrower frequency range of interest and next   \n"
"load the big data.                                                            \n"
"   You may also choose to modify the data: if high frequencies do not matter, \n"
"data could be binned into coarser time grid. Alternatively, you may split data\n"
"into shorter intervals and take average of several periodograms, thus keeping \n"
"the frequency range while degrading resolution.                               \n"
"   This kind of problems are fundamental in time-frequency analysis and depend\n" 
"on no algorithm. Violation of the uncertainity principle Dt*Df<1, where either\n" 
"Dt-time span and Df-frequency resolution or Dt-resolution and Df-span yields  \n" 
"respectively undersampling or clipping of (effective) Nyquist interval. \n");
    };
    free(d); free(ind);
    return cre_spec(nf,fr0,frs);
}
/*-----------------------------------------------------------*/

void peak (LONG n, FLOAT fx[], TIME * xm, FLOAT * fm, FLOAT * dx)
/*
 ! Scan periodogram for a peak
 ! INPUT:
 ! n- periodogram length
 ! fx-periodogram values
 ! OUTPUT:
 ! xm-peak location
 ! fm-peak value
 ! dx-peak halfwidth
 ! SPECIAL CONDITIONS:
 ! fm<0 no peak in valid region
 ! dx<0 no valid (parabolic) peak
 ! METHOD:
 ! For f=log(fx) fits a parabola 0.5(d2*x**2+d1*x)+f(2)
 ! where d2=f(1)+f(3)-2f(2), d1=f(3)-f(1)
 ! finds dxl & dxp such, that the linear and quadratic terms drop
 ! by 0.7 in log, i.e. by a factor of 2 (dx=HWHI)
 !(C) Alex Schwarzenberg-Czerny, 1999-2005 alex@camk.edu.pl
 */
{
    INT i, nm, n0;
    FLOAT f[3], d2, d1;
    /*
     * valid maximum?
     */
    d1 = fx[0];
    nm=0;
    for (i = 1; i < n; i++)
        if (fx[i] > d1) {
            d1 = fx[i];
            nm = i;
        };
    *fm = (FLOAT) d1;
    *xm = nm;
    *dx = - (FLOAT) 1.;
    if (*fm < (FLOAT) 0.) //linear case (peak at edge)
        return;
    n0 = MAX (1, MIN (n - 2, nm) );
    for (i = 0; i < 3; i++)
        f[i] = (FLOAT) log (MAX (fabs (fx[i + n0 - 1]) / (*fm), EPSILON) );
    d1 = f[2] - f[0];
    *dx = (FLOAT) (-fabs (d1) / 1.4);
    /* parabolic case (not on edge) */
    d2 = - (f[0] - f[1] - f[1] + f[2]);
    if (d2 > (FLOAT) 0.7 * (*dx) * (*dx) ) {
        /* original 1.4 */
        *fm = (FLOAT) exp (0.125 * d1 * d1 / d2 + f[1]) * (*fm);
        *xm = 0.5 * d1 / d2 + n0;
        *dx = (FLOAT) sqrt (0.7 * d2);
    }
    *dx = (FLOAT) 1. / (*dx);
}

void peakrm(TIME dt, SPEC_T spec,
            TIME *freq, FLOAT *thm, FLOAT *dfm) {
// ! Removes a peak from the spectrum to search for subsequent ones.
    INT nfm,iwdt,j,ni,nx;

    peak (spec.nfr, spec.th, freq, thm, dfm);
    *dfm=*dfm*spec.frs;
    nfm = (INT) (*freq + 0.5);
    *freq = *freq * spec.frs + spec.fr0;
    iwdt =  (INT) (1.7 / fabs(dt*(FLOAT)spec.frs))+1;
/* remove peaks +-iwdt */
    ni = MAX (0, nfm - iwdt);
    nx = MIN (spec.nfr, nfm + iwdt + 1);
    for (j = ni; j < nx; j++)
                spec.th[j] = - (FLOAT) fabs (spec.th[j]);
}



