/* gcc -o m mySort.c -lm */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "time.h"

#define FLOAT double
#define INT int
#define FFMT "%f "
#define IFMT "%d "
/*--------------sort_merge------------------------*/
/*
 * global variables and subroutines for sort_merge
 * external FLOAT defined as float or double, as required
 */
static INT *ina, *inb;
static FLOAT *a;

static void merge_part (INT k, INT m, INT l) {
    INT i, j, r;
    for (j = k; j <= m; j++)
        inb[j - k] = ina[j];
    i = 0;
    r = k;
    while (r < j && j <= l)
        if (a[inb[i]] <= a[ina[j]])
            ina[r++] = inb[i++];
        else
            ina[r++] = ina[j++];
    while (r < j)
        ina[r++] = inb[i++];
}


static void sort_part (INT k, INT l) {
    INT m;
    if (k < l) {
        m = (k + l) / 2;
        sort_part (k, m);
        sort_part (m + 1, l);
        merge_part (k, m, l);
    }
}


void sortx_m (INT n, FLOAT f[], INT ind[])
/*
 * Purpose:
 Sort of index IND with key A, both of length N
 Method: Stable,
 Merge sort, recurrence implementation,
 requires extra memory (inb[(N+1)/2]).
 Alex Schwarzenberg-Czerny, 2005 alex@camk.edu.pl
 */
{
    INT i;
    a = f;
    ina = ind;
//    inb = malloc ( (n + 1) / 2, sizeof (INT) );
    inb = malloc ( (n + 1) / 2 * sizeof (INT) );
    for (i = 0; i < n; i++)
        ina[i] = i;
    sort_part (0, n - 1);
    free (inb);
}

//=================mergeSort========================

void mergeSort(INT n, FLOAT *a, INT *i, INT *s)
/* Returns index 'i' sorting a[i[N]], s[N] is a work array,
   N is size of all arrays. 'a' remains intact. 
   Method: merge sort, this implementation declares no internal arrays,
   though ~10log_2(N) local variables are put on stack.
   Original free code from wiki talk, adapted to be stable and to perform
   index sort by
   Alex Schwarzenberg-Czerny, 2018, alex@camk.edu.pl
*/
{
  if(n<=1)
    return;
  {
    INT j = 0;
    INT mid = n/2;
    INT* left  = i;
    INT* right = &i[mid];
    INT* endr  = &i[n];
    INT* endl  = right;
    /* Sort sub arrays */
    mergeSort(mid, a, left, s);
    mergeSort(n-mid, a, right, s);
    /* Merge sorted sub arrays */
    for(j=0; j<n; j++){
      if(left < endl && (a[*left]<=a[*right] || right>=endr))
        s[j] = *left++;
      else
        s[j] = *right++;    
    }
    for(j=0; j<n; j++)
      i[j] = s[j];
  }
}
//=================test code========================

static void test_sort(INT N, FLOAT *x) {
  /* Check result */
  INT i;
  FLOAT t;
  t=x[0];
  for (i=1;i<N;i++)
    t+=x[i];
  for (i=1;i<N;i++) {
    if (x[i]<x[i-1])
      break;
  }
  if (i<N)
    printf("Not sorted, sum=%g\n",t);
  else
    printf("Sorted, sum=%g\n",t);
}

static void test_stability(INT N, FLOAT *x, INT *ind) {
  INT i,stable=1;
  for (i=1;i<N;i++) {
    if (x[i]==x[i-1])
      if (ind[i]<ind[i-1]) {
        stable=0;
        printf("Not stable\n");
        return;
      };
  };
  printf("Stable\n");
}

//====================Main==========================
/* Example of how to call the sort algorithm */
#define N 100000000
#define PRN 0

int main(void)
{
  INT i,opt=1;
  FLOAT *x; /* Input vector */
  FLOAT *s; /* Scratch space */
  INT *ind,*is;
  clock_t start, end;   /* start and end times */

  /* no other memory requirements */
  x = malloc(sizeof(FLOAT)*(N));
  s = malloc(sizeof(FLOAT)*(N));
  ind = malloc(sizeof(INT)*(N));
  is = malloc(sizeof(INT)*(N));
  /* Test sort performance
     ...get random array */
  srand ( 1949 );
  //srand ( time(NULL) );
  for(i=0;i<N;i++){
    ind[i] = i;
    x[i] = rand() / (float) RAND_MAX;
    if (PRN) printf(IFMT"%f ",i,x[i]);
  }
  printf("N= "IFMT"\n",N);

  test_sort(N, x);
  /* Call sort algorithm */
  switch( opt ) 
  {
    case 1:
        {start = clock();
         printf("mergesort\n"); // Time: 44.156154 [sec]
         mergeSort(N, x, ind, is);
         end = clock();
         for (i=0;i<N;i++)
           s[i]=x[ind[i]];
         memcpy(x,s,sizeof(FLOAT)*(N));
        }
        break;
    case 2:
        {start = clock();
         printf("sortx_m\n"); // Time: 42.003307 [sec]
         sortx_m(N, x, ind);
         end = clock();
         for (i=0;i<N;i++)
           s[i]=x[ind[i]];
         memcpy(x,s,sizeof(FLOAT)*(N));
        }
        break;
    default :
        {printf("\nwrong case\n");}
        break;
  }  
  test_sort(N, x);
  printf("Time: %Lf [sec]\n", (long double)(end-start)/CLOCKS_PER_SEC);

  // stability test 
  for(i=0;i<N;i++){
    ind[i]=i;
    x[i] = 5.5;
    if (PRN) printf(IFMT"%f ",i,x[i]);
  }
  printf("N= "IFMT"\n",N);

  switch( opt ) 
  {
    case 1:
        {start = clock();
         printf("mergesort\n"); // Time: 44.156154 [sec]
         mergeSort(N, x, ind, is);
         end = clock();
         for (i=0;i<N;i++)
           s[i]=x[ind[i]];
         memcpy(x,s,sizeof(FLOAT)*(N));
        }
        break;
    case 2:
        {start = clock();
         printf("sortx_m\n"); // Time: 42.003307 [sec]
         sortx_m(N, x, ind);
         end = clock();
         for (i=0;i<N;i++)
           s[i]=x[ind[i]];
         memcpy(x,s,sizeof(FLOAT)*(N));
        }
        break;
    default :
        {printf("\nwrong case\n");}
        break;
  }
 
  test_stability(N, x, ind);
  free(ind);
  free(s);
  free(x);
  return 0;
}

